// The "hitbox" and size of each character sprite.
int characterWidth = 60;
int characterHeight = 60;

/* Tracks data for a character in the game. */
class CharacterData {
  PImage sprite;
  int health;
  int maxHealth;
  int moveLeft;
  int maxMove;
  boolean isDead;
  int x;
  int y;
  boolean isActive;  // True if this is the character the player is controlling right now.
  RoomData curRoom;  // For reference, the RoomData structure representing where the character currently sits.
  
  CharacterData(PartyMemberInfo _info) {
    sprite = loadImage(_info.spritePicture);
    maxHealth = _info.maxHealth;
    health = maxHealth;
    maxMove = _info.maxMovement;
    moveLeft = maxMove;
    isDead = false;
    x = 0;
    y = 0;
    isActive = false;
    curRoom = null;
  }
  
  /* Draws the appropriate character sprite. Assumes that translation for scrolling has _not_ been performed. */
  void draw() {
    if(isDead) {
      tint(128);
      image(deadSprite, x + scrollX, y + scrollY);
    } else {
      tint(255);
      image(sprite, x + scrollX, y + scrollY);
    }
  }
  
  /* Puts the character in a random position in the given room. */
  void putInRoom(RoomData room) {
    int minX = room.unitsX * sizeUnitScale + characterWidth;
    int maxX = (room.unitsX + room.roomRef.unitWidth) * sizeUnitScale - characterWidth;
    int minY = room.unitsY * sizeUnitScale + characterHeight;
    int maxY = (room.unitsY + room.roomRef.unitHeight) * sizeUnitScale - characterHeight;
    x = int(random(minX, maxX));
    y = int(random(minY, maxY));
    curRoom = room;
  }
  
  /* Puts the character in the given room next to the given door. */
  void putInRoomByDoor(RoomData room, DoorInfo door) {
    curRoom = room;
    // TODO: Replace this with real stuff.
    x = room.unitsX * sizeUnitScale + (room.roomRef.unitWidth * sizeUnitScale / 2);
    y = room.unitsY * sizeUnitScale + (room.roomRef.unitHeight * sizeUnitScale / 2);
  }
  
  /* Increase or decrease the character's health by the given value, react accordingly. */
  void changeHealth(int healthDelta) {
    health += healthDelta;
    if(health < 1) {
      isDead = true;
    } else {
      isDead = false;
    }
  }
  
  /* Increase or decrease the character's movement points by the given value, react accordingly. 
     Returns true if the value of moveLeft actually changed.*/
  boolean changeMovement(int moveDelta) {
    int oldMove = moveLeft;
    moveLeft += moveDelta;
    if(moveLeft < 0) {
      moveLeft = 0;
    } else if(moveLeft > maxMove) {
      moveLeft = maxMove;
    }
    return oldMove != moveLeft;
  }
  
  /* Try to shift the character's position by the given values.
     Returns true if the character actually moved. */
  boolean shift(int _xShift, int _yShift) {
    boolean moved = false;
    boolean madeNewRoom = false;  // Used to track potential event kick-offs.
    int collideDoor = -1;
    if(moveLeft > 0) {
      x += _xShift;
      y += _yShift;
      // Check for a door collision.
      collideDoor = checkDoorCollision();
      if(collideDoor >= 0) {
        // Try to find a valid existing room on the other side of the door or make a new one.
        int newRoomIndex = -1;
        int newDoorIndex = -1;
        newRoomIndex = getNeighborRoom(curRoom, curRoom.roomRef.doors.get(collideDoor));
        if(newRoomIndex >= 0) {
          // Confirm that the room on the other side of the door has a connecting door.
          newDoorIndex = getNextRoomDoor(curRoom, discoveredRooms.get(newRoomIndex), curRoom.roomRef.doors.get(collideDoor));
          if(newDoorIndex >= 0) {
            curRoom = discoveredRooms.get(newRoomIndex);
            putInRoomByDoor(curRoom, curRoom.roomRef.doors.get(newDoorIndex));
            moved = true;
          }
        } else {
          // If there's not a neighboring room yet, make one.
            newRoomIndex = generateNewRoom(curRoom, curRoom.roomRef.doors.get(collideDoor));
            if(newRoomIndex >= 0) {
              // Confirm that the room on the other side of the door has a connecting door.
              newDoorIndex = getNextRoomDoor(curRoom, discoveredRooms.get(newRoomIndex), curRoom.roomRef.doors.get(collideDoor));
              if(newDoorIndex >= 0) {
                curRoom = discoveredRooms.get(newRoomIndex);
                putInRoomByDoor(curRoom, curRoom.roomRef.doors.get(newDoorIndex));
                moved = true;
              } else {
                // NOTE: Weird state, made a new (supposedly valid) room, but no connecting door found. If it comes up often, handle appropriately.
              }
              madeNewRoom = true;
            } else {
              // NOTE: Weird state, making a new room failed. If this comes up often, add code to handle appropriately.
            }
        }
        if(madeNewRoom) {
          // TODO: This is where "discovered a new room" events would potentially kick off. Call a function to see if one happens, and then execute one if so.
        }
      } else {  
        // Check for a wall collision if no door collision.
        if(checkWallCollision()) {
          x -= _xShift;
          y -= _yShift;
        } else {
          moved = true;
        }
      }
      // If, after all that, the character actually moved, decrement their move points.
      if(moved) {
        changeMovement(-1);
      }
    }
    return moved;
  }
  
  /* Checks for a collision between the character and any wall of the room it's in. */ 
  boolean checkWallCollision() {
    boolean collision = false;
    if(curRoom != null) {
      int roomMinX = curRoom.unitsX * sizeUnitScale;
      int roomMaxX = (curRoom.unitsX + curRoom.roomRef.unitWidth) * sizeUnitScale;
      int roomMinY = curRoom.unitsY * sizeUnitScale;
      int roomMaxY = (curRoom.unitsY + curRoom.roomRef.unitHeight) * sizeUnitScale;
      if(x < roomMinX || x + characterWidth > roomMaxX || y < roomMinY || y + characterHeight > roomMaxY) {
        collision = true;
      }
    }
    return collision;
  }
  
  /* Checks if the character has "collided" with any of the doors in its current room.
     Returns -1 if no, or the index into the room's doors array of the door that collided. */
  int checkDoorCollision() {
    int collidedDoor = -1;
    if(curRoom != null) {
      collidedDoor = curRoom.checkCharacterInDoor(this);
    }
    return collidedDoor;
  }
}

/* Adjusts the active character's position by the given values. */
void moveActiveCharacter(int _xShift, int _yShift) {
  for(int i = 0; i < characters.size(); i++) {
    if(characters.get(i).isActive && characters.get(i).moveLeft > 0) {
      characterMoved = characters.get(i).shift(_xShift, _yShift);
    }
  }
}

/* Finds and returns the index of the first active character, or -1 if there are none. */
int getActiveCharacterIndex() {
  for(int i = 0; i < characters.size(); i++) {
    if(characters.get(i).isActive) {
      return i;
    }
  }
  return -1;
}
