// How many pixels to shift a door along its respective wall segment.
int doorOffset = 150;
int[] doorWidths = {40, 40, 20, 20};
int[] doorHeights = {70, 5, 40, 40};
int doorHitboxPadding = 3;  // How many pixels to extend the "hit box" of room doors.
int wallHeight = 80;

/* Class to hold information about "discovered" rooms. */
class RoomData {
  PImage floorImage;
  PImage wallImage;
  // The top-left position of the room in grid units.
  int unitsX;
  int unitsY;
  ArrayList<DoorInfo> doors;
  RoomInfo roomRef;
  
  RoomData(RoomInfo _room, int _unitsX, int _unitsY) {
    floorImage = loadImage(_room.floorPicture);
    wallImage = loadImage(_room.wallPicture);
    unitsX = _unitsX;
    unitsY = _unitsY;
    doors = _room.doors;
    roomRef = _room;
  }
  
  /* Assumes translation for scrolling has _not_ been done. */
  void drawFloor() {
    // Draw the floor.
    if(isACharacterInRoom(this)) {
      tint(255);
    } else {
      tint(128);
    }
    image(floorImage, unitsX * sizeUnitScale + scrollX, unitsY * sizeUnitScale + scrollY);
  }
  
  /* Draw the walls and doors separately so they can end up "on top" of all other scenery.
     Assumes translation for scrolling has _not_ been done. */
  void drawWallAndDoors() {
    // Draw the wall and door(s) for the room. Separated so it can be done after all floors are drawn.
    if(isACharacterInRoom(this)) {
      tint(255);
    } else {
      tint(128);
    }
    image(wallImage, unitsX * sizeUnitScale + scrollX, unitsY * sizeUnitScale - wallHeight + scrollY);
    // Draw any doors.
    for(int i = 0; i < doors.size(); i++) {
      int offset = doors.get(i).offset;
      switch(doors.get(i).side) {
        case 0: // North.
          // The "wallHeight - doorHeights..." compensates for the door not necessarily being as tall as the wall.
          image(doorN, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
        case 1: // South.
          image(doorS, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
        case 2: // East.
          image(doorE, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
        case 3: // West.
          image(doorW, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
      }
    }
  }
  
  /* Checks if the given character is "in" any of this room's doors.
     Returns the index of the door in the doors array, or -1 if they're not colliding with any doors. */
  int checkCharacterInDoor(CharacterData character) {
    int doorIndex = -1;
    int doorMinX, doorMaxX, doorMinY, doorMaxY; 
    int charMinX = character.x;
    int charMaxX = character.x + characterWidth;
    int charMinY = character.y;
    int charMaxY = character.y + characterWidth;
    for(int i = 0; i < doors.size(); i++) {
      DoorInfo curDoor = doors.get(i);
      int offset = curDoor.offset;
      doorMinX = getDoorX(curDoor) - doorHitboxPadding;
      doorMaxX = doorMinX + doorWidths[curDoor.side] + doorHitboxPadding;
      doorMinY = getDoorY(curDoor) - doorHitboxPadding;
      doorMaxY = doorMinY + doorHeights[curDoor.side] + doorHitboxPadding;
      // Check if "in" x-wise, then y-wise. If both, it's a collision.
      if((charMinX <= doorMaxX && charMinX >= doorMinX) || (charMaxX >= doorMinX && charMaxX <= doorMaxX) || (charMinX <= doorMinX && charMaxX >= doorMaxX)) {
        if((charMinY <= doorMaxY && charMinY >= doorMinY) || (charMaxY >= doorMinY && charMaxY <= doorMaxY) || (charMinY <= doorMinY && charMaxY >= doorMaxY)) {
          doorIndex = i;
          break;
        }
      }
    }
    return doorIndex;
  }
  
  /* Returns the X coordinate of the given door. */
  int getDoorX(DoorInfo door) {
    int offset = door.offset;
    int doorX = 0;
    switch(door.side) {
      case 0: // North.
        doorX = (unitsX + offset) * sizeUnitScale + doorOffset - (doorWidths[door.side] / 2); 
        break;
      case 1: // South.
        doorX = (unitsX + offset) * sizeUnitScale + doorOffset - (doorWidths[door.side] / 2);
        break;
      case 2: // East.
        doorX = (unitsX + roomRef.unitWidth) * sizeUnitScale - doorWidths[door.side];
        break;
      case 3: // West.
        doorX = unitsX * sizeUnitScale;
        break;
    }
    return doorX;
  }
  
  /* Returns the Y coordinate of the given door. */
  int getDoorY(DoorInfo door) {
    int offset = door.offset;
    int doorY = 0;
    switch(door.side) {
      case 0: // North.
        doorY = unitsY * sizeUnitScale - wallHeight + (wallHeight - doorHeights[door.side]); 
        break;
      case 1: // South.
        doorY = (unitsY + roomRef.unitHeight) * sizeUnitScale - doorHeights[door.side];
        break;
      case 2: // East.
        doorY = (unitsY + offset) * sizeUnitScale + doorOffset - (doorHeights[door.side] / 2);
        break;
      case 3: // West.
        doorY = (unitsY + offset) * sizeUnitScale + doorOffset - (doorHeights[door.side] / 2);
        break;
    }
    return doorY;
  }
}

/* Randomly selects a RoomInfo to create a new room from.
   If _needsDoorDirection is -1, any room will do. Otherwise,
   requires the room have a N (0), S (1), E (2), or W (2) door. */
RoomInfo getNewRoom(int _needsDoorDirection) {
  boolean acceptable = false;
  RoomInfo room = roomLoader.getRoomInfo(int(random(roomLoader.size())));;
  while(!acceptable) {
    if(_needsDoorDirection >= 0) {
      for(DoorInfo door : room.doors) {
        if(door.side == _needsDoorDirection) {
          acceptable = true;
        }
      }
    } else {
      // No door requirements, so any room will do.
      acceptable = true;
    }
    if(!acceptable) {
      room = roomLoader.getRoomInfo(int(random(roomLoader.size()))); 
    }
  }
  return room;
}

/* Checks if any character is in a given room. */
boolean isACharacterInRoom(RoomData room) {
  for(int i = 0; i < characters.size(); i++) {
    int charUnitX = int(characters.get(i).x / sizeUnitScale);
    int charUnitY = int(characters.get(i).y / sizeUnitScale);
    
    if(!characters.get(i).isDead && characters.get(i).curRoom == room) {
      return true;
      /*
      if(charUnitX >= room.unitsX && charUnitX <= room.unitsX + room.roomRef.unitWidth) {
        if(charUnitY >= room.unitsY && charUnitY <= room.unitsY + room.roomRef.unitHeight) {
          return true;
        }
      }
      */
    }
  }
  return false;
}

/* Attempts to find a room next to curRoom through the door at doorIndex.
   Returns the index of that room, or -1 if none exists yet. */
int getNeighborRoom(RoomData curRoom, DoorInfo curDoor) {
  int neighborRoomIndex = -1;
  for(int i = 0; i < discoveredRooms.size(); i++) {
    RoomData checkRoom = discoveredRooms.get(i);
    switch(curDoor.side) {
      case 0: // North.
        if(curRoom.unitsY == checkRoom.unitsY + checkRoom.roomRef.unitHeight) {
          if(curRoom.unitsX + curDoor.offset >= checkRoom.unitsX && curRoom.unitsX + curDoor.offset <= checkRoom.unitsX + checkRoom.roomRef.unitWidth) {
            return i;
          }
        }
        break;
      case 1: // South.
        if(curRoom.unitsY + curRoom.roomRef.unitHeight == checkRoom.unitsY) {
          if(curRoom.unitsX + curDoor.offset >= checkRoom.unitsX && curRoom.unitsX + curDoor.offset <= checkRoom.unitsX + checkRoom.roomRef.unitWidth) {
            return i;
          }
        }
        break;
      case 2: // East.
        if(curRoom.unitsX + curRoom.roomRef.unitWidth == checkRoom.unitsX) {
          if(curRoom.unitsY + curDoor.offset >= checkRoom.unitsY && curRoom.unitsY + curDoor.offset <= checkRoom.unitsY + checkRoom.roomRef.unitHeight) {
            return i;
          }
        }
        break;
      case 3: // West.
        if(curRoom.unitsX == checkRoom.unitsX + checkRoom.roomRef.unitWidth) {
          if(curRoom.unitsY + curDoor.offset >= checkRoom.unitsY && curRoom.unitsY + curDoor.offset <= checkRoom.unitsY + checkRoom.roomRef.unitHeight) {
            return i;
          }
        }
        break;
    }
  }
  return neighborRoomIndex;
}

/* Attempts to find an existing room next to the given room through the given door,
   or generate an appropriate new one. Returns the index of the new room in the
   discoveredRooms array, or -1 if a room couldn't be found/generated. */
int generateNewRoom(RoomData curRoom, DoorInfo curDoor) {
  int newRoomIndex = -1;
  int neededDoorSide = -1;
  int newX = 0;
  int newY = 0;
  
  switch(curDoor.side) {
    case 0: // North.
      neededDoorSide = 1;
      break;
    case 1: // South.
      neededDoorSide = 0;
      break;
    case 2: // East.
      neededDoorSide = 3;
      break;
    case 3:
      neededDoorSide = 2;
      break;
  }
  RoomInfo newRoomInfo = getNewRoom(neededDoorSide);
  // Set the new room's X and Y based on which door side is connecting to the current room.
  // TODO: WARNING: Assumption made here that all rooms are 2 units wide by 2 units high.
  switch(curDoor.side) {
    case 0: // New room is North.
      newX = curRoom.unitsX;
      newY = curRoom.unitsY - newRoomInfo.unitHeight;
      break;
    case 1: // New room is South.
      newX = curRoom.unitsX;
      newY = curRoom.unitsY + curRoom.roomRef.unitHeight;
      break;
    case 2: // New room is East.
      newX = curRoom.unitsX + curRoom.roomRef.unitWidth;
      newY = curRoom.unitsY;
      break;
    case 3: // New room is West.
      newX = curRoom.unitsX - newRoomInfo.unitWidth;
      newY = curRoom.unitsY;
      break;
  }
  RoomData newRoom = new RoomData(newRoomInfo, newX, newY);
  discoveredRooms.add(newRoom);
  newRoomIndex = discoveredRooms.size() - 1;
  return newRoomIndex;
}

/* Given an "exiting" room & door and the desired "entering" room,
   tries to find the appropriate door in the new room that connects to the
   door of the old room. Returns the index of the door for the new room, or
   -1 if there isn't a matching door. */
int getNextRoomDoor(RoomData curRoom, RoomData nextRoom, DoorInfo curDoor) {
  int newDoorIndex = -1;
  ArrayList<DoorInfo> doors = nextRoom.roomRef.doors;
  DoorInfo checkDoor;
  int doorX = 0;
  int doorY = 0;
  int neededSide = -1;
  int checkDoorX = 0;
  int checkDoorY = 0;
  
  switch(curDoor.side) {
    case 0: // North.
      doorX = curRoom.unitsX + 2 * curDoor.offset;
      doorY = curRoom.unitsY;
      neededSide = 1;
      break;
    case 1: // South.
      doorX = curRoom.unitsX + 2 * curDoor.offset;
      doorY = curRoom.unitsY + curRoom.roomRef.unitHeight;
      neededSide = 0;
      break;
    case 2: // East.
      doorX = curRoom.unitsX + curRoom.roomRef.unitWidth;
      doorY = curRoom.unitsY + 2 * curDoor.offset;
      neededSide = 3;
      break;
    case 3: // West.
      doorX = curRoom.unitsX;
      doorY = curRoom.unitsY + 2 * curDoor.offset;
      neededSide = 2;
      break;
  }
  // Look through the doors of the target room trying to find a match.
  for(int i = 0; i < doors.size(); i++) {
    checkDoor = doors.get(i);
    if(checkDoor.side == neededSide) {
      switch(checkDoor.side) {
        case 0: // North.
          checkDoorX = nextRoom.unitsX + 2 * curDoor.offset;
          checkDoorY = nextRoom.unitsY;
          break;
        case 1: // South.
          checkDoorX = nextRoom.unitsX + 2 * curDoor.offset;
          checkDoorY = nextRoom.unitsY + nextRoom.roomRef.unitHeight;
          break;
        case 2: // East.
          checkDoorX = nextRoom.unitsX + nextRoom.roomRef.unitWidth;
          checkDoorY = nextRoom.unitsY + 2 * curDoor.offset;
          break;
        case 3: // West.
          checkDoorX = nextRoom.unitsX;
          checkDoorY = nextRoom.unitsY + 2 * curDoor.offset;
          break;
      }
      if(doorX == checkDoorX && doorY == checkDoorY) {
        newDoorIndex = i;
        break;
      }
    }
  }
  return newDoorIndex;
}
