int hudWidth = 100;
int nameBoxFontSize = 16;
int nameBoxInternalPadding = 4;  // Padding between the text in each name box and its border.
int nameBoxWidth = hudWidth;
int nameBoxHeight = nameBoxFontSize + (nameBoxInternalPadding * 2);
color nameBoxFill = color(128, 255, 255);
color nameBoxBorder = nameBoxFill;
color nameBoxTextColor = color(0, 0, 0);
int portraitWidth = hudWidth;
int portraitHeight = 100;
int healthBarWidth = hudWidth;
int healthBarHeight = 10;
int movementBarWidth = hudWidth;
int movementBarHeight = 10;
int maxMovementPoints = 100;  // The most movement points any character can have.
int hudElementPadding = 3;  // Vertical space between HUD elements.
int hudHeight = nameBoxHeight + portraitHeight + healthBarHeight + movementBarHeight + (3 * hudElementPadding);

color mouseOverColor = color(0, 255, 0, 64);
color selectedColor = color(128, 255, 128, 64);

int hudLeft = 1150;  // The x of the character HUDs.
int hudTop = 20;  // The initial top of the first character HUD.
int hudPadding = 10;  // Space between each character HUD.
color hudBorderColor = color(0, 255, 255);

/* Represents the Heads-Up Display for a character. */
class CharacterHUD {
  int x;
  int y;
  int w;  // Used for "is the mouse over me?" checks.
  int h;  // Used for "is the mouse over me?" checks.
  TextBox nameBox;
  CharacterPortrait portrait;
  HealthBar healthBar;
  MovementBar moveBar;
  String characterName;
  CharacterData characterRef;  // The "real" data associated with this character.
  
  CharacterHUD(int _x, int _y, PartyMemberInfo _info, CharacterData _char) {
    x = _x;
    y = _y;
    w = hudWidth;
    h = hudHeight;
    characterName = _info.name;
    characterRef = _char;
    
    nameBox = new TextBox(x, y, nameBoxWidth, nameBoxHeight, nameBoxFontSize, nameBoxBorder, nameBoxFill);
    portrait = new CharacterPortrait(x, y + nameBoxHeight + hudElementPadding, _info);
    healthBar = new HealthBar(x, y + nameBoxHeight + portraitHeight + (hudElementPadding * 2), _info.maxHealth, _info.maxHealth);
    moveBar = new MovementBar(x, y + nameBoxHeight + portraitHeight + healthBarHeight + (hudElementPadding * 3), _info.maxMovement, _info.maxMovement);
  }
  
  /* Sets the display value for the health of a character. Also adjusts portrait picture. */
  void setHealth(int newHealth) {
    healthBar.current = newHealth;
    if(newHealth < healthBar.max) {
      portrait.isHurt = true;
    } else {
      portrait.isHurt = false;
    }
    if(newHealth < 1) {
      portrait.isDead = true;
    } else {
      portrait.isDead = false;
    }
  }
  
  /* Sets the display value for the movement left of a character. */
  void setMovement(int newMovement) {
    moveBar.current = newMovement;
  }
  
  /* Draws each HUD element. */
  void draw() {
    setHealth(characterRef.health);
    setMovement(characterRef.moveLeft);
    nameBox.displayText(characterName, nameBoxTextColor);
    portrait.draw();
    healthBar.draw();
    moveBar.draw();
    // Draw extra stuff if the mouse is over the HUD or it's selected.
    if(isMouseOver() == true && characterRef.isActive == false) {
      noStroke();
      fill(mouseOverColor);
      rect(x - (hudPadding / 2), y - (hudPadding / 2), w + hudPadding, h + hudPadding);
    } else if(characterRef.isActive) {
      noStroke();
      fill(selectedColor);
      rect(x - (hudPadding / 2), y - (hudPadding / 2), w + hudPadding, h + hudPadding);
    }
  }
  
  /* Checks if the mouse is over any part of this HUD element. */
  boolean isMouseOver() {
    return (mouseX >= x && mouseX <= x + w && mouseY >= y && mouseY <= y + h);
  }
}

/* Represents a character's portrait as part of the "HUD" for that character. */
class CharacterPortrait {
  int x;
  int y;
  int w;
  int h;
  PImage healthyImage;
  PImage hurtImage;
  PImage deadImage;
  boolean isHurt;
  boolean isDead;
  
  CharacterPortrait(int _x, int _y, PartyMemberInfo _info) {
    x = _x;
    y = _y;
    w = portraitWidth;
    h = portraitHeight;
    isHurt = false;
    isDead = false;
    healthyImage = loadImage(_info.healthyPicture);
    hurtImage = loadImage(_info.hurtPicture);
    deadImage = loadImage(_info.deadPicture);
  }
  
  void draw() {
    tint(255);
    if(isDead == true) {
      image(deadImage, x, y, w, h);
    } else if(isHurt == true) {
      image(hurtImage, x, y, w, h);
    } else {
      image(healthyImage, x, y, w, h);
    }
  }
}

/* Represents a health bar to display under a character's portrait. */
class HealthBar {
  int x;
  int y;
  int w;
  int h;
  int max;  // Maximum health points.
  int current;  // Current health points.
  // Size of each "tick" on the health bar.
  int tickWidth;
  int tickHeight;
  int tickSpace;
  color borderCol;
  color healthCol;
  color hurtCol;
  
  HealthBar(int _x, int _y, int _max, int _current) {
    x = _x;
    y = _y;
    w = healthBarWidth;
    h = healthBarHeight;
    max = _max;
    current = _current;
    
    tickWidth = 30;
    tickHeight = 10;
    tickSpace = 5;
    borderCol = color(128, 0, 0);
    healthCol = color(255, 0, 0);
    hurtCol = color(0, 0, 0);
  }
  
  /* Draw a series of health "ticks". */
  void draw() {
    stroke(borderCol);
    for(int i = 0; i < max; i++) {
      if(i + 1 <= current) {
        fill(healthCol);
      } else {
        fill(hurtCol);
      }
      rect(x + (i * (tickWidth + tickSpace)), y, tickWidth, tickHeight);
    }
  }
}

/* Represents a "movement left" bar to display under a character's portrait. */
class MovementBar {
  int x;
  int y;
  int w;
  int h;
  int max;  // Maximum movement points.
  int current;  // Current movement points.
  color borderCol;
  color barCol;  // The color to put in the "you have this much left" part of the bar.
  color usedBarCol;  // The color to put in the "you've used this much" part of the bar.
  int barPadding;  // How much space to put between between the border and the bar.

  MovementBar(int _x, int _y, int _max, int _current) {
    x = _x;
    y = _y;
    w = movementBarWidth;
    h = movementBarHeight;
    max = _max;
    current = _current;
    
    borderCol = color(255, 255, 255);
    barCol = color(255, 255, 0);
    usedBarCol = color(0, 0, 0);
  }
  
  /* Draws a horizontal colored bar representing movement left. */
  void draw() {
    stroke(borderCol);
    fill(usedBarCol);
    int baseBarWidth = int(float(max) / float(maxMovementPoints) * w);
    int barWidth = int((float(current) / float(max)) * baseBarWidth);
    if(barWidth < 0) { barWidth = 0; }
    if(barWidth > w) { barWidth = w; }
    rect(x, y, baseBarWidth, h);
    fill(barCol);
    rect(x + barPadding, y + barPadding, barWidth - (barPadding * 2), h - (barPadding * 2));
  }
}


